// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.CommandLine
open System.CommandLine.Invocation
open System.IO
open System.Threading.Tasks

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let WriteTempCsproj (directory: string) (dotnetVersionString: string) : unit =
    let csprojPath = Path.Combine(directory, $"Project_{dotnetVersionString}.csproj")

    let csprojContent =
        $"""<Project Sdk="Microsoft.NET.Sdk">
  <PropertyGroup>
    <NoWarn>\$(NoWarn);NETSDK1138</NoWarn>
    <RollForward>Major</RollForward>
    <TargetFramework>net{dotnetVersionString}</TargetFramework>
  </PropertyGroup>
</Project>"""

    use writer = new StreamWriter(csprojPath)

    do writer.Write(csprojContent)
    writer.Flush()

let GetSdkVersionChain (min: double) (max: double) : float list =
    let generated: List<float> = [ System.Math.Ceiling(min) .. max ]

    float min :: float max :: generated |> List.distinct |> List.sort

// TODO: "gdmt-restore" never fails, use heuristic to determine if it has failed?
let RunGdmtRestore (args: List<string>) : unit =
    ExecProcess("gdmt-restore" :: args).Run().Check()

let CommandName = "gdmt-checkcore"
let CommandDescription = "check versions of NuGet pkgs bound to SDK version"

let SdkMinVersionOption =
    new Option<string>([| "-m"; "--sdk-min" |], "minimal SDK version")

SdkMinVersionOption.SetDefaultValue "6.0"

let SdkMaxVersionOption =
    new Option<string>([| "-e"; "--sdk-ver" |], "targeted SDK version")

SdkMaxVersionOption.SetDefaultValue "8.0"

let SdkExecutableOption =
    new Option<string>([| "-x"; "--sdk-exe" |], "SDK executable path")

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let sdkMinVersion = double (options.GetValueForOption SdkMinVersionOption)
                let sdkMaxVersion = double (options.GetValueForOption SdkMaxVersionOption)

                let tempGuid = Guid.NewGuid().ToString()
                let tempName = $"gdmt_check_core_{tempGuid}"
                let tempPath = Path.Combine(Path.GetTempPath(), tempName)

                SimpleLog.LogMessage SimpleLog.Debug $"Temp path is: {tempPath}"

                for sdkVersion in GetSdkVersionChain sdkMinVersion sdkMaxVersion do
                    let sdkVersionString = sprintf "%.1f" sdkVersion
                    let sdkVersionTempPath = Path.Combine(tempPath, sdkVersionString)

                    SimpleLog.LogMessage SimpleLog.Debug $"Starting restore for .NET {sdkVersionString}"

                    Directory.CreateDirectory sdkVersionTempPath |> ignore
                    WriteTempCsproj sdkVersionTempPath sdkVersionString

                    let mutable gdmtRestoreArgs = [ "-p"; sdkVersionTempPath ]

                    match options.GetValueForOption SdkExecutableOption with
                    | emptyValue when String.IsNullOrEmpty emptyValue -> ()
                    | nonEmptyValue -> gdmtRestoreArgs <- List.append gdmtRestoreArgs [ "--sdk-exe"; nonEmptyValue ]

                    RunGdmtRestore gdmtRestoreArgs
                    Directory.Delete(sdkVersionTempPath, true)

                Directory.Delete(tempPath, true)

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddOption SdkMinVersionOption
    rootCommand.AddOption SdkMaxVersionOption
    rootCommand.AddOption SdkExecutableOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
