// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.SimpleLog

let UnwantedFileGlobs =
    [ "CoreFxPipe_PSHost.*"
      "MSBuild*"
      "NuGetScratch*"
      "System.IO.Directory System.Runtime*"
      "System.IO.File System.Runtime*"
      "clr-debug-pipe-*"
      "dotnet-diagnostic-*"
      "system-commandline-sentinel-files" ]

let DirectoryDeleteRecursive (path: string) : unit = Directory.Delete(path, true)

let CleanUnwantedFilesFromTmpdir (path: string) (dryRun: bool) : unit =
    for unwantedFileGlob in UnwantedFileGlobs do
        let unwantedFiles =
            Directory.GetDirectories(path, unwantedFileGlob)
            |> Array.append (Directory.GetFiles(path, unwantedFileGlob))

        for unwantedFile in unwantedFiles do
            let (fileType: string, removalFunction: string -> unit) =
                match unwantedFile with
                | d when Directory.Exists d -> ("dir ", DirectoryDeleteRecursive)
                | _ -> ("file", File.Delete)

            SimpleLog.LogMessage SimpleLog.Debug $"Discovered {fileType}: {unwantedFile}"

            if not dryRun then
                removalFunction unwantedFile
                SimpleLog.LogMessage SimpleLog.Success $"Successfully removed {fileType}: {unwantedFile}"

let CommandName = "gdmt-cleantmp"
let CommandDescription = "clean .NET artifacts from a given tmpdir"

let TmpdirArgument = new Argument<string>("tmpdir", "tmpdir to clean")
TmpdirArgument.SetDefaultValue "/tmp"

let DryRunOption = new Option<bool>([| "-d"; "--dry-run" |], "dry run")
DryRunOption.SetDefaultValue false

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let tmpdir = options.GetValueForArgument TmpdirArgument
                let dryRun = options.GetValueForOption DryRunOption

                CleanUnwantedFilesFromTmpdir tmpdir dryRun

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddArgument TmpdirArgument

    rootCommand.AddOption DryRunOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
