// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Text.RegularExpressions
open System.Threading.Tasks

open LibGit2Sharp
open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.PackageName
open Gdmt.Shared.SimpleLog

let GetChangedFiles (repoTopDirectory: string) : string list =
    use repo = new Repository(repoTopDirectory)

    repo.Diff.Compare<TreeChanges>(repo.Head.Tip.Tree, DiffTargets.WorkingDirectory)
    |> Seq.map (fun change -> Path.GetFileName change.Path)
    |> List.ofSeq

let UpdateEbuildManifests () : unit =
    SimpleLog.LogMessage SimpleLog.Debug "Updating manifests"
    ExecProcess([ "pkgdev"; "manifest" ]).Run().Check()
    SimpleLog.LogMessage SimpleLog.Success "Successfully updated manifests"

let GitAddAll () : unit =
    ExecProcess([ "git"; "add"; "--all"; "." ]).Run().Check()

let PkgcheckScan () : unit =
    SimpleLog.LogMessage SimpleLog.Debug "Scanning with pkgcheck"
    ExecProcess([ "pkgcheck"; "scan"; "--exit"; "error" ]).Run().Check()
    SimpleLog.LogMessage SimpleLog.Success "Scan with pkgcheck succeeded"

let IsEbuildFile (path: string) : bool = Regex.IsMatch(path, ".*\.ebuild$")

let IsLiveEbuildFile (path: string) : bool =
    IsEbuildFile path && Regex.IsMatch(path, ".*9999\..*")

let GetLatestNonLiveEbuildFile (path: string) : string =
    let latestEbuilds =
        Directory.GetFiles(path, "*.ebuild")
        |> Array.filter (IsLiveEbuildFile >> not)
        |> Array.sort
        |> Array.rev

    match latestEbuilds with
    | [||] -> "no latest ebuild file available" |> Exception |> raise
    | _ -> latestEbuilds |> Array.head

let TestEbuilds (ebuildFiles: seq<string>) : unit =
    for ebuild: string in ebuildFiles do
        match ebuild with
        | f when File.Exists f ->
            SimpleLog.LogMessage SimpleLog.Debug $"Testing ebuild: {f}"

            ExecProcess([ "gdmt-ebuild"; f; "clean"; "install"; "clean" ]).Run().Check()

            SimpleLog.LogMessage SimpleLog.Success $"Ebuild test passed for: {f}"
        | f -> SimpleLog.LogMessage SimpleLog.Debug $"Ebuild file does not exist: {f}"

let GuardCommitMessage (commitMessage: string) : unit =
    if String.IsNullOrEmpty commitMessage then
        "the effective commit message is empty" |> Exception |> raise

let CreateTags (tagType: string) (sources: string[]) : string =
    sources
    |> Array.map (fun source ->
        match source with
        | s when String.IsNullOrEmpty s -> ""
        | s when Regex.IsMatch(s, "http.*") -> $"{tagType}: {s}\n"
        | s -> $"{tagType}: https://bugs.gentoo.org/{s}\n")
    |> String.concat ""

let CommandName = "gdmt-commit"
let CommandDescription = "commit ebuild packages"

// Git / commit message

let DryRunOption = new Option<bool>([| "-d"; "--dry" |], "do not make commits")
DryRunOption.SetDefaultValue false

let GitCommitMessageOption =
    new Option<string>([| "-m"; "--message" |], "commit message")

let GitCommitMessageAuxOption =
    new Option<string>([| "-M"; "--message-aux" |], "auxiliary commit message")

let GitSignOption = new Option<bool>([| "-S"; "--sign" |], "sign the create commit")

GitSignOption.SetDefaultValue false

// Tests

let ScanOption = new Option<bool>([| "-s"; "--scan" |], "scan using pkgcheck")
ScanOption.SetDefaultValue false

let TestOption = new Option<bool>([| "-t"; "--test" |], "test using ebuild")
TestOption.SetDefaultValue false

// Bump

let BumpOption =
    new Option<string>([| "-B"; "--bump" |], "set message to \"bump to VERSION\"")

let BumpCopyOption =
    new Option<bool>([| "-C"; "--copy" |], "if doing a bump copy latest ebuild")

BumpCopyOption.SetDefaultValue false

// Drop

let DropOption =
    new Option<string>([| "-D"; "--drop" |], "set message to \"drop old VERSION\"")

let DropRemoveOption =
    new Option<bool>([| "-R"; "--remove" |], "if doing a drop, remove ebuild")

DropRemoveOption.SetDefaultValue false

// Sync live

let SyncLiveOption =
    new Option<bool>([| "-L"; "--live" |], "copy latest non-9999 into 9999 ebuild")

SyncLiveOption.SetDefaultValue false

// Bugs

let BugOption =
    new Option<string>([| "-b"; "--bug" |], "reference a Gentoo BugZilla or upstream bug/URL")

let ClosesOption =
    new Option<string>([| "-c"; "--closes" |], "closes a Gentoo BugZilla bug")

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let dryRun = options.GetValueForOption DryRunOption

                let gitCommitMessage = options.GetValueForOption GitCommitMessageOption
                let gitCommitMessageAux = options.GetValueForOption GitCommitMessageAuxOption
                let gitSign = options.GetValueForOption GitSignOption

                let scan = options.GetValueForOption ScanOption
                let test = options.GetValueForOption TestOption

                let bump = options.GetValueForOption BumpOption
                let bumpCopy = options.GetValueForOption BumpCopyOption
                let drop = options.GetValueForOption DropOption
                let dropRemove = options.GetValueForOption DropRemoveOption
                let syncLive = options.GetValueForOption SyncLiveOption

                let bugSources = options.GetValueForOption BugOption
                let closesSources = options.GetValueForOption ClosesOption

                let mutable commitMessage = ""

                // CATEGORY/PN
                let packageFullName = PackageName.GetFullPackageName "."
                // PN
                let packageShortName = Path.GetFileName packageFullName

                if not (String.IsNullOrEmpty bump) then
                    commitMessage <- $"{packageFullName}: bump to {bump}"

                    if bumpCopy then
                        let latest = GetLatestNonLiveEbuildFile "."
                        let updated = $"{packageShortName}-{bump}.ebuild"

                        SimpleLog.LogMessage SimpleLog.Debug $"Copying {latest} into {updated}"
                        File.Copy(latest, updated)

                        if not dryRun then
                            ExecProcess([ "git"; "add"; updated ]).Run().Check()

                if not (String.IsNullOrEmpty drop) then
                    commitMessage <- $"{packageFullName}: drop old {drop}"

                    if dropRemove then
                        let old = $"{packageShortName}-{drop}.ebuild"

                        SimpleLog.LogMessage SimpleLog.Debug $"Removing {old}"
                        File.Delete(old)

                if syncLive then
                    commitMessage <- $"{packageFullName}: sync live 9999 version"

                    let latest = GetLatestNonLiveEbuildFile "."
                    let updated = $"{packageShortName}-9999.ebuild"

                    SimpleLog.LogMessage SimpleLog.Debug $"Copying {latest} into {updated}"
                    File.Copy(latest, updated, true)

                    if not dryRun then
                        ExecProcess([ "git"; "add"; updated ]).Run().Check()

                let changes = GetChangedFiles "../.."
                let changedEbuilds = changes |> List.filter IsEbuildFile

                for change in changes do
                    SimpleLog.LogMessage SimpleLog.Debug $"Detected change: {change}"

                for change in changedEbuilds do
                    SimpleLog.LogMessage SimpleLog.Debug $"Modified ebuild: {change}"

                UpdateEbuildManifests()

                if not dryRun then
                    GitAddAll()

                if scan then
                    PkgcheckScan()

                if not (String.IsNullOrEmpty gitCommitMessage) then
                    commitMessage <- $"{packageFullName}: {gitCommitMessage}"

                GuardCommitMessage commitMessage

                if test then
                    match changedEbuilds with
                    | [] -> SimpleLog.LogMessage SimpleLog.Debug "No changed ebuilds, nothing to do"
                    | _ -> TestEbuilds changedEbuilds

                let bugTags =
                    match bugSources with
                    | s when String.IsNullOrEmpty s -> ""
                    | s -> CreateTags "Bug" (s.Split [| ' ' |])

                let closesTags =
                    match closesSources with
                    | s when String.IsNullOrEmpty s -> ""
                    | s -> CreateTags "Closes" (s.Split [| ' ' |])

                let commitMessageFull =
                    commitMessage
                    + "\n"
                    + match gitCommitMessageAux with
                      | s when String.IsNullOrEmpty s -> ""
                      | s -> "\n" + s + "\n"
                    + "\n"
                    + bugTags
                    + closesTags

                if not dryRun then
                    let gitCommitArgs =
                        [ "git"; "commit"; "-m"; $"'{commitMessageFull}'" ]
                        @ (if gitSign then [ "--signoff" ] else [])

                    ExecProcess(gitCommitArgs).Run().Check()

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddOption DryRunOption

    rootCommand.AddOption GitCommitMessageOption
    rootCommand.AddOption GitCommitMessageAuxOption
    rootCommand.AddOption GitSignOption

    rootCommand.AddOption ScanOption
    rootCommand.AddOption TestOption

    rootCommand.AddOption BumpOption
    rootCommand.AddOption BumpCopyOption
    rootCommand.AddOption DropOption
    rootCommand.AddOption DropRemoveOption
    rootCommand.AddOption SyncLiveOption

    rootCommand.AddOption BugOption
    rootCommand.AddOption ClosesOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
