// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let CommandName = "gdmt-ebuild"

let CommandDescription =
    "wrapper to run the ebuild command faster and for multiple files"

// Ebuild command

let EbuildPathArgument =
    new Argument<string>("ebuild-path", "ebuild file path or directory of ebuilds")

let EbuildCommandArgument =
    new Argument<string[]>("ebuild-command", "ebuild command(s), e.g. prepare, compile")

EbuildCommandArgument.SetDefaultValue [| "manifest"; "clean"; "install" |]

// Package manager (Portage)

let PortageFeaturesOption =
    new Option<string>([| "-f"; "--features" |], "Portage features")

PortageFeaturesOption.SetDefaultValue "test"

let PortageUseOption = new Option<string>([| "-u"; "--use" |], "Portage USE flags")
PortageUseOption.SetDefaultValue "test"

let ForceTestsOption = new Option<bool>([| "-t"; "--test" |], "force running tests")
ForceTestsOption.SetDefaultValue false

// System stress

let NicenessOption = new Option<int>([| "-n"; "--nice" |], "niceness level")
NicenessOption.SetDefaultValue 19

let Nproc = Environment.ProcessorCount

let JobsOption = new Option<int>([| "-j"; "--jobs" |], "max number of jobs")
JobsOption.SetDefaultValue Nproc

let LoadAverageOption =
    new Option<int>([| "-l"; "--load" |], "max number of load average")

LoadAverageOption.SetDefaultValue Nproc

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let ebuildPath = options.GetValueForArgument(EbuildPathArgument)

                let ebuildCommand =
                    options.GetValueForArgument(EbuildCommandArgument) |> Array.toList

                let portageFeatures = options.GetValueForOption(PortageFeaturesOption)
                let portageUse = options.GetValueForOption(PortageUseOption)
                let forceTests = options.GetValueForOption(ForceTestsOption)

                let niceness = options.GetValueForOption(NicenessOption)
                let jobs = options.GetValueForOption(JobsOption)
                let loadAverage = options.GetValueForOption(LoadAverageOption)

                let ebuilds =
                    match ebuildPath with
                    | p when Directory.Exists p -> Directory.GetFiles(ebuildPath, "*.ebuild")
                    | p when File.Exists p -> [| p |]
                    | _ -> $"The ebuild file does not exist, given {ebuildPath}" |> Exception |> raise

                let portageTestFlag = if forceTests then " test" else ""

                let envMap =
                    [ ("ALLOW_TEST", (if forceTests then "all" else ""))
                      ("FEATURES", portageFeatures + portageTestFlag)
                      ("MAKEOPTS", $"-j {jobs} -l {loadAverage}")
                      ("USE", portageUse + portageTestFlag) ]

                for ebuild in ebuilds do
                    let args = [ "nice"; "-n"; $"{niceness}"; "ebuild"; ebuild ] @ ebuildCommand

                    SimpleLog.LogMessage SimpleLog.Debug $"Running ebuild command for {ebuild}"
                    ExecProcess(args, envMap).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success $"Ebuild command for {ebuild} finished successfully"

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription


    rootCommand.AddArgument EbuildPathArgument
    rootCommand.AddArgument EbuildCommandArgument

    rootCommand.AddOption PortageFeaturesOption
    rootCommand.AddOption PortageUseOption
    rootCommand.AddOption ForceTestsOption

    rootCommand.AddOption NicenessOption
    rootCommand.AddOption JobsOption
    rootCommand.AddOption LoadAverageOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
