// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let CommandName = "gdmt-gensdk"
let CommandDescription = "maintenance tool to create a .NET SDK distfile"

let SdkVersionArgument = new Argument<string>("sdk-version", ".NET SDK version")

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let sdkVersion = options.GetValueForArgument SdkVersionArgument

                let tempGuid = Guid.NewGuid().ToString()
                let tempName = $"gdmt_gensdk_{tempGuid}"

                let tempPath = Path.Combine(Path.GetTempPath(), tempName)
                let tempSdkPath = Path.Combine(tempPath, $"dotnet-sdk-{sdkVersion}")

                let tempCache = Path.Combine(tempPath, "cache")
                let tempEnvTempdir = Path.Combine(tempCache, "tmpdir")

                let sdkDistTarballName = $"dotnet-sdk-{sdkVersion}.tar.xz"
                let sdkDistTarballPath = Path.Combine(tempPath, sdkDistTarballName)

                let numThreads = Environment.ProcessorCount
                let numSafeThreads = if numThreads > 2 then numThreads - 2 else 1

                let envMap =
                    [ ("TMPDIR", tempEnvTempdir)
                      ("XZ_OPT", $"-T{numSafeThreads}")

                      ("DOTNET_CLI_TELEMETRY_OPTOUT", "1")
                      ("DOTNET_NOLOGO", "1")
                      ("DOTNET_SKIP_FIRST_TIME_EXPERIENCE", "1")
                      ("MSBUILDDISABLENODEREUSE", "1")

                      ("UseSharedCompilation", "false") ]

                for (envKey, envVar) in envMap do
                    Environment.SetEnvironmentVariable(envKey, envVar, EnvironmentVariableTarget.Process)

                // Info

                SimpleLog.LogMessage SimpleLog.Debug $"Temp path:      {tempPath}"
                SimpleLog.LogMessage SimpleLog.Debug $"Temp SDK path:  {tempSdkPath}"
                SimpleLog.LogMessage SimpleLog.Debug $"TEMPDIR env:    {tempEnvTempdir}"

                // Clone the official .NET builder project repository.

                Directory.CreateDirectory tempEnvTempdir |> ignore

                SimpleLog.LogMessage SimpleLog.Debug $"Running repository clone"
                Environment.CurrentDirectory <- tempPath

                let sdkCloneArgs =
                    [ "git"
                      "clone"
                      "--branch"
                      $"v{sdkVersion}"
                      "--depth"
                      "1"
                      "--recursive"
                      "--shallow-submodules"
                      "https://github.com/dotnet/dotnet"
                      tempSdkPath ]

                ExecProcess(sdkCloneArgs).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success $"Repository clone succeeded"

                // Run "prep.sh".

                SimpleLog.LogMessage SimpleLog.Debug "Running script prep.sh"
                Environment.CurrentDirectory <- tempSdkPath

                ExecProcess([ "bash"; "prep.sh" ]).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success $"Script run succeeded"

                // Cleanup.

                SimpleLog.LogMessage SimpleLog.Debug "Cleaning up"

                ExecProcess([ "rm"; "-f"; "-r"; ".git" ]).Run().Check()

                // Tar up.

                SimpleLog.LogMessage SimpleLog.Debug $"Creating tarball {sdkDistTarballName}"
                Environment.CurrentDirectory <- tempPath

                let sdkTarArgs =
                    [ "tar"
                      "--create"
                      "--auto-compress"
                      "--file"
                      sdkDistTarballPath
                      $"dotnet-sdk-{sdkVersion}" ]

                ExecProcess(sdkTarArgs).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success $"Created tarball at path {sdkDistTarballPath}"

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddArgument SdkVersionArgument

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
