// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let CreateArchive (inputDirectory: string) (outputFile: string) : unit =
    if File.Exists outputFile then
        File.Delete outputFile

    let numThreads = Environment.ProcessorCount
    let numSafeThreads = if numThreads > 2 then numThreads - 2 else 1

    let args =
        [ "tar"; "--auto-compress"; "--create"; "--file"; outputFile; inputDirectory ]

    let env = [ ("XZ_OPT", $"-T{numSafeThreads}") ]

    SimpleLog.LogMessage SimpleLog.Debug $"Creating archive from:  {inputDirectory}"
    SimpleLog.LogMessage SimpleLog.Debug $"Creating archive file:  {outputFile}"

    ExecProcess(args, env).Run().Check()

let CommandName = "gdmt-govendor"
let CommandDescription = "create a Go dependency/vendor archive"

let ProjectDirectoryArgument =
    new Argument<string>("project-directory", "Go project directory to use")

ProjectDirectoryArgument.SetDefaultValue "."

let OutputPrefixOption =
    new Option<string>([| "-o"; "--output" |], "target archive output path prefix")

let OutputVersionOption =
    new Option<string>([| "-r"; "--outver" |], "version appended to output prefix")

let GoArchiveTypeOption =
    new Option<string>([| "-t"; "--type" |], "Go archive type (dependency/vendor)")

GoArchiveTypeOption.SetDefaultValue "dependency"

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let projectDirectory =
                    options.GetValueForArgument ProjectDirectoryArgument |> Path.GetFullPath

                let projectName = Path.GetFileName projectDirectory

                let outputPrefix = options.GetValueForOption OutputPrefixOption
                let outputVersion = options.GetValueForOption OutputVersionOption

                let goArchiveType = options.GetValueForOption GoArchiveTypeOption

                Environment.CurrentDirectory <- projectDirectory

                let outputPrefixPath =
                    let outputPrefixNamed =
                        match outputPrefix with
                        | p when String.IsNullOrEmpty p -> $"./{projectName}"
                        | p -> p

                    match outputVersion with
                    | v when String.IsNullOrEmpty v -> outputPrefixNamed
                    | v -> $"{outputPrefixNamed}-{v}"

                SimpleLog.LogMessage SimpleLog.Debug $"Project directory:  {projectDirectory}"
                SimpleLog.LogMessage SimpleLog.Debug $"Project name:       {projectName}"
                SimpleLog.LogMessage SimpleLog.Debug $"Go archive type:    {goArchiveType}"
                SimpleLog.LogMessage SimpleLog.Debug $"Output prefix:      {outputPrefixPath}"

                let archivePath =
                    match goArchiveType with
                    | "dependency" ->
                        let archive = $"{outputPrefixPath}-deps.tar.xz"

                        let goModCache = Path.Combine(projectDirectory, "go-mod")
                        let goArgs = [ "go"; "mod"; "download"; "-modcacherw"; "-x" ]
                        let goEnv = [ ("GOMODCACHE", goModCache) ]

                        ExecProcess(goArgs, goEnv).Run().Check()

                        CreateArchive "go-mod" archive

                        Path.GetFullPath archive
                    | "vendor" ->
                        let archive = $"{outputPrefixPath}-vendor.tar.xz"
                        let vendorPath = (Path.Combine(projectName, "vendor"))

                        ExecProcess([ "go"; "mod"; "vendor" ]).Run().Check()

                        Environment.CurrentDirectory <- ".."
                        CreateArchive vendorPath archive

                        Path.GetFullPath archive
                    | _ -> $"no known Go archive path, given {goArchiveType}" |> Exception |> raise

                SimpleLog.LogMessage SimpleLog.Success $"Successfully created archive: {archivePath}"

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddArgument ProjectDirectoryArgument

    rootCommand.AddOption OutputPrefixOption
    rootCommand.AddOption OutputVersionOption
    rootCommand.AddOption GoArchiveTypeOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
