// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.CommandLine
open System.CommandLine.Invocation
open System.IO
open System.Threading.Tasks

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let LinuxRuntimes: string list =
    [
      // generic
      "linux"

      // glibc
      "linux-arm"
      "linux-arm64"
      "linux-x64"
      "linux-x86"

      // musl
      "linux-musl-arm"
      "linux-musl-arm64"
      "linux-musl-x64"
      "linux-musl-x86" ]

let DotnetRuntimeNugets: string list =
    [ "microsoft.aspnetcore.app.ref"
      "microsoft.aspnetcore.app.runtime."
      "microsoft.netcore.app.host."
      "microsoft.netcore.app.ref"
      "microsoft.netcore.app.runtime." ]

let FindSdkExecutable (sdkVersion: string) : string =
    let srcExe = $"/usr/bin/dotnet-{sdkVersion}"
    let binExe = $"/usr/bin/dotnet-bin-{sdkVersion}"

    if File.Exists(srcExe) then
        srcExe
    elif File.Exists(binExe) then
        binExe
    else
        "no suitable dotnet sdk executable was found" |> Exception |> raise

let GetFileDirectory (path: string) =
    match path with
    | d when Directory.Exists d -> d
    | f when File.Exists f -> Path.GetDirectoryName(f)
    | _ -> $"path {path} does not exist" |> Exception |> raise

let RestoreForRuntime
    (sdkExe: string)
    (runtime: string)
    (projectPath: string)
    (tmpDir: string)
    (nugetPackagesDir: string)
    (extraRestoreArgs: string)
    : bool =

    let env =
        [ ("PWD", GetFileDirectory projectPath)
          ("TMPDIR", tmpDir)
          ("NUGET_PACKAGES", nugetPackagesDir)

          ("DOTNET_CLI_TELEMETRY_OPTOUT", "1")
          ("DOTNET_NOLOGO", "1")
          ("DOTNET_SKIP_FIRST_TIME_EXPERIENCE", "1")
          ("MSBUILDDISABLENODEREUSE", "1")
          ("UseSharedCompilation", "false") ]

    let args =
        [ sdkExe
          "restore"
          "--force-evaluate"
          "--verbosity"
          "quiet"
          "--runtime"
          runtime
          extraRestoreArgs
          projectPath ]

    let execProcess = ExecProcess(args, env)

    execProcess.Run().Success

let ForEachRuntime (lambdaForRuntime: string -> bool) : unit =
    LinuxRuntimes
    |> List.map (fun runtime ->
        SimpleLog.LogMessage SimpleLog.Debug $"Performing restore for runtime: {runtime}"

        let exitSuccess = lambdaForRuntime runtime

        (runtime, exitSuccess))
    |> List.map (fun (runtime, exitSuccess) ->
        match exitSuccess with
        | true -> SimpleLog.LogMessage SimpleLog.Success $"Restore for runtime {runtime} passed successfully"
        | _ -> SimpleLog.LogMessage SimpleLog.Error $"Restore for runtime {runtime} has failed")
    |> ignore

let GatherNugets (nugetPackagesDir: string) : List<string> =
    try
        Directory.GetDirectories(nugetPackagesDir, "*", SearchOption.AllDirectories)
        |> Seq.map (fun path -> path.Substring(nugetPackagesDir.Length + 1))
        |> Seq.map (fun path -> path.Split('/'))
        |> Seq.filter (fun pathElements -> pathElements.Length = 2)
        |> Seq.map (fun pathElements -> $"{pathElements.[0]}@{pathElements.[1]}")
        |> Seq.filter (fun atNuget -> not (DotnetRuntimeNugets |> List.exists atNuget.Contains))
        |> Seq.sort
        |> Seq.toList
    with ex ->
        SimpleLog.LogMessage SimpleLog.Error ex.Message
        []

let CommandName = "gdmt-restore"
let CommandDescription = "maintenance tool to restore projects"

let ProjectPathOption = new Option<string>([| "-p"; "--project" |], "project path")
ProjectPathOption.SetDefaultValue "."

let SdkVersionOption = new Option<string>([| "-e"; "--sdk-ver" |], "SDK version")
SdkVersionOption.SetDefaultValue "8.0"

let SdkExecutableOption =
    new Option<string>([| "-x"; "--sdk-exe" |], "SDK executable path")

let CacheDirOption =
    new Option<string>([| "-c"; "--cache" |], "cache directory to use")

let ExtraRestoreArgsOption =
    new Option<string>([| "-r"; "--extra-r" |], "extra restore arguments")

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let projectPath = options.GetValueForOption(ProjectPathOption) |> Path.GetFullPath

                if not (File.Exists(projectPath) || Directory.Exists(projectPath)) then
                    $"project {projectPath} does not exist" |> Exception |> raise

                let cacheDir =
                    match options.GetValueForOption(CacheDirOption) with
                    | emptyValue when String.IsNullOrEmpty emptyValue ->
                        Path.Combine(projectPath |> GetFileDirectory, ".cache")
                    | nonEmptyValue -> nonEmptyValue |> Path.GetFullPath

                let tmpDir = Path.Combine(cacheDir, "tmpdir")
                let nugetPackagesDir = Path.Combine(cacheDir, "nuget_packages")

                let sdkExecutable =
                    match options.GetValueForOption(SdkExecutableOption) with
                    | emptyValue when String.IsNullOrEmpty emptyValue ->
                        SdkVersionOption |> options.GetValueForOption |> FindSdkExecutable
                    | nonEmptyValue -> nonEmptyValue

                let extraRestoreArgs = ExtraRestoreArgsOption |> options.GetValueForOption

                Directory.CreateDirectory tmpDir |> ignore

                SimpleLog.LogMessage SimpleLog.Debug $"SDK executable path:       {sdkExecutable}"
                SimpleLog.LogMessage SimpleLog.Debug $"Project path:              {projectPath}"
                SimpleLog.LogMessage SimpleLog.Debug $"Cache directory:           {cacheDir}"
                SimpleLog.LogMessage SimpleLog.Debug $"TMPDIR directory:          {tmpDir}"
                SimpleLog.LogMessage SimpleLog.Debug $"NUGET_PACKAGES directory:  {nugetPackagesDir}"

                ForEachRuntime(fun runtime ->
                    RestoreForRuntime sdkExecutable runtime projectPath tmpDir nugetPackagesDir extraRestoreArgs)

                printfn "Those are the gathered NuGet packages:"
                printfn ""

                GatherNugets nugetPackagesDir
                |> List.map (fun nuget -> printfn "%s" nuget)
                |> ignore

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddOption ProjectPathOption
    rootCommand.AddOption SdkVersionOption
    rootCommand.AddOption SdkExecutableOption
    rootCommand.AddOption CacheDirOption
    rootCommand.AddOption ExtraRestoreArgsOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
