// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Net.Http
open System.Text.RegularExpressions
open System.Threading.Tasks

open FSharp.Data
open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

type StabilizationArrayProvider =
    JsonProvider<"""
    [
        {
            "category": "Category",
            "package": "Package",
            "version": "Version",
            "message": "Message"
        }
    ]
""">

let StabilizationUriBase = "https://packages.gentoo.org/maintainer/"

let GetStabilizationArray (maintainerEmail: string) =
    async {
        let stabilizationUri =
            $"{StabilizationUriBase}/{maintainerEmail}/stabilization.json"

        let httpClient = new HttpClient()
        let! httpResponse = httpClient.GetStringAsync(stabilizationUri) |> Async.AwaitTask
        let jsonResult = StabilizationArrayProvider.Parse(httpResponse)

        return jsonResult
    }

let GetStabilizationArchitectures (messageString: string) =
    messageString.Split([| ' ' |])
    |> Array.filter (fun s -> s[0] = '~')
    |> Array.map (fun s -> s.Replace("~", "").Replace(",", ""))

let AskYesNoRegex = new Regex(@"^[Yy]([Ee][Ss])?$")

let AskYesNo () =
    match Console.ReadLine() with
    // Return (Enter) key was pressed.
    | "" -> true
    | s -> AskYesNoRegex.IsMatch(s)

let PackageWithRevisionRegex = new Regex(@".+-r[0-9]+$")

let RunPkgdevBugs (arches: string) (package: string) =
    // When there is no revision specified we can use the "~" package operator
    // in order to say "match any revision", if there is a revision specified,
    // then we have to be exact and use "=" package operator.
    let adequatePackageAtom =
        match package with
        | s when PackageWithRevisionRegex.IsMatch(s) -> $"={package}"
        | _ -> $"~{package}"

    let pkgdevBugsArgs =
        [ "pkgdev"; "bugs"; "--stablereq"; "--arches"; arches; adequatePackageAtom ]

    let success = ExecProcess(pkgdevBugsArgs).Run().Success

    if success then
        SimpleLog.LogMessage SimpleLog.Success $"Successfully submitted request for {package}"
    else
        SimpleLog.LogMessage SimpleLog.Error $"Errors occurred when processing request"
        SimpleLog.LogMessage SimpleLog.Error $"Failed to submit request for {package}"

let ProcessStabilizationItem (stabilizationItem: StabilizationArrayProvider.Root) =
    let category = stabilizationItem.Category
    let package = stabilizationItem.Package
    let version = stabilizationItem.Version
    let message = stabilizationItem.Message

    let packageFullName = $"{category}/{package}-{version}"
    let arches = GetStabilizationArchitectures message |> String.concat ","

    SimpleLog.LogMessage SimpleLog.Debug $"Full package: {packageFullName}"
    SimpleLog.LogMessage SimpleLog.Debug $"Message: {message}"
    SimpleLog.LogMessage SimpleLog.Debug $"Stabilization architectures: {arches}"

    printf $"Use the package {packageFullName}? [Y/n] "

    if AskYesNo() then
        RunPkgdevBugs arches packageFullName

let CommandName = "gdmt-stablereq"
let CommandDescription = "file stabilization requests on Gentoo Bugzilla"

let MaintainerArgument =
    new Argument<string>("maintainer", "target maintainer e-mail to query package.gentoo.org")

let RepositoryOption =
    new Option<string>([| "-r"; "--repo" |], "ebuild repository to use, if given location changes to it")

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let maintainer = options.GetValueForArgument(MaintainerArgument)
                let repository = options.GetValueForOption(RepositoryOption)

                match repository with
                | empty when String.IsNullOrEmpty empty -> ()
                | repo when Directory.Exists(repo) ->
                    SimpleLog.LogMessage SimpleLog.Success $"Using ebuild repository {repo}"
                    Environment.CurrentDirectory <- repo
                | repo ->
                    SimpleLog.LogMessage SimpleLog.Error $"Ebuild repository does not exist, given {repo}"
                    $"no such directory {repo}" |> Exception |> raise

                let! stabilizationArray = GetStabilizationArray maintainer

                for stabilizationItem in stabilizationArray do
                    ProcessStabilizationItem stabilizationItem

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddArgument MaintainerArgument

    rootCommand.AddOption RepositoryOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
