// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let CommandName = "gdmt-sync"
let CommandDescription = "synchronize the ebuild repository"

// Remotes

let RemotesArgument = new Argument<string[]>("remotes", "remotes to push to")
RemotesArgument.SetDefaultValue [| "origin" |]

// Git

let GitPullFromRemoteOption =
    Option<string>([| "-f"; "--from" |], "pull from specified remote")

GitPullFromRemoteOption.SetDefaultValue "origin"

let GitSignOrigin =
    Option<bool>([| "-s"; "--sign" |], "sign when pushing to origin remote")

GitSignOrigin.SetDefaultValue false

// System stress

let Nproc = Environment.ProcessorCount

let JobsOption = new Option<int>([| "-j"; "--jobs" |], "max number of jobs")
JobsOption.SetDefaultValue Nproc

let LoadAverageOption =
    new Option<int>([| "-l"; "--load" |], "max number of load average")

LoadAverageOption.SetDefaultValue Nproc

let CommandHandler =
    Func<InvocationContext, Task>(fun context ->
        let operation =
            async {
                let options = context.ParseResult

                let remotes = options.GetValueForArgument RemotesArgument

                let gitPullFromRemote = options.GetValueForOption GitPullFromRemoteOption
                let gitSignOrigin = options.GetValueForOption GitSignOrigin

                let jobs = options.GetValueForOption JobsOption
                let loadAverage = options.GetValueForOption LoadAverageOption

                SimpleLog.LogMessage SimpleLog.Debug $"Scanning with pkgcheck"
                ExecProcess([ "pkgcheck"; "scan"; "--commits"; "--exit=error" ]).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success "Scan succeeded"

                SimpleLog.LogMessage SimpleLog.Debug $"Pulling from remote {gitPullFromRemote}"
                ExecProcess([ "git"; "pull"; gitPullFromRemote ]).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success $"Pull from {gitPullFromRemote} succeeded"

                if gitSignOrigin && Array.contains "origin" remotes then
                    SimpleLog.LogMessage SimpleLog.Debug "Pushing to remote: origin"
                    ExecProcess([ "git"; "push"; "origin"; "--sign" ]).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success "Push to origin succeeded"

                for remote in remotes do
                    if gitSignOrigin && remote = "origin" then
                        ()
                    else
                        SimpleLog.LogMessage SimpleLog.Debug $"Pushing to remote: {remote}"
                        ExecProcess([ "git"; "push"; remote ]).Run().Check()
                        SimpleLog.LogMessage SimpleLog.Success $"Push to {remote} succeeded"

                let egencacheArgs =
                    [ "egencache"
                      "--jobs"
                      $"{jobs}"
                      "--load-average"
                      $"{loadAverage}"
                      "--repo"
                      "gentoo"
                      "--update"
                      "--update-pkg-desc-index"
                      "--update-use-local-desc" ]

                SimpleLog.LogMessage SimpleLog.Debug "Updating ebuild cache"
                ExecProcess(egencacheArgs).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success "Cache update succeeded"

                SimpleLog.LogMessage SimpleLog.Debug "Updating eix database"
                ExecProcess([ "eix-update"; "--quiet" ]).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success "Database update succeeded"

                SimpleLog.LogMessage SimpleLog.Debug "Refreshing the git repository index"
                ExecProcess([ "git"; "update-index"; "--refresh" ]).Run().Check()
                SimpleLog.LogMessage SimpleLog.Success "Refresh succeeded"

                SimpleLog.LogMessage SimpleLog.Debug "Resolving updates"
                ExecProcess([ "eselect"; "news"; "list"; "new" ]).Run().Check()

                let eixUpdateArgs = [ "eix"; "--upgrade"; "--only-names" ]
                ExecProcess(eixUpdateArgs, [ ("EIX_LIMIT", "0") ]).Run().Check()

                ()
            }

        Async.StartAsTask(operation, TaskCreationOptions.None))

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(CommandName)

    rootCommand.Name <- CommandName
    rootCommand.Description <- CommandDescription

    rootCommand.AddArgument RemotesArgument

    rootCommand.AddOption GitPullFromRemoteOption
    rootCommand.AddOption GitSignOrigin

    rootCommand.AddOption JobsOption
    rootCommand.AddOption LoadAverageOption

    rootCommand.SetHandler CommandHandler

    rootCommand.Invoke(argv)
